import pytest

from inuits_policy_based_auth import BaseAuthenticationPolicy
from inuits_policy_based_auth.contexts import UserContext
from inuits_policy_based_auth.exceptions import (
    AuthenticateMethodDidNotReturnObjectOfTypeUserContextException,
)
from unittest.mock import Mock, patch


class TestBaseAuthenticationPolicy:
    @patch(
        "inuits_policy_based_auth.authentication.base_authentication_policy.BaseAuthenticationPolicy.__abstractmethods__",
        set(),
    )
    def setup_method(self, _):
        self.policy = BaseAuthenticationPolicy()  # type: ignore
        self.spy_policy_authenticate = Mock()
        self.user_context = UserContext()

        self.policy.authenticate = self.spy_policy_authenticate

    def test_apply_returns_user_context(self):
        self.policy.authenticate.return_value = self.user_context

        user_context = self.policy.apply(self.user_context)

        self.policy.authenticate.assert_called_once()
        assert user_context is self.user_context

    def test_apply_raises_AuthenticateMethodDidNotReturnObjectOfTypeUserContextException(
        self,
    ):
        with pytest.raises(
            AuthenticateMethodDidNotReturnObjectOfTypeUserContextException
        ):
            self.policy.apply(self.user_context)

        self.policy.authenticate.assert_called_once()



class RequestContext:
    """A class containing data about the context of a request.

    The data provided in this class is used in policies to determine whether a
    user is authenticated/authorized.

    Properties
    ----------
    http_request : Unknown
        An HTTP request used by policies to determine access to a resource.
    resource_scopes : list[str], optional
        Scopes a user must have for accessing a protected resource when policies
        rely on scope-based authorization.
    """

    def __init__(self, http_request, resource_scopes: list[str] = []):
        """
        Parameters
        ----------
        http_request : Unknown
            An HTTP request used by policies to determine access to a resource.
        resource_scopes : list[str], optional
            Scopes a user must have for accessing a protected resource when policies
            rely on scope-based authorization.
        """

        self._http_request = http_request
        self._resource_scopes = resource_scopes

    @property
    def http_request(self):
        """
        An HTTP request used by policies to determine access to a resource.
        """

        return self._http_request

    @property
    def resource_scopes(self):
        """
        Scopes a user must have for accessing a protected resource when policies
        rely on scope-based authorization.
        """

        return self._resource_scopes



class AccessRestrictions:
    """
    A class containing properties used to restrict access.

    The value of the properties of this class are defined by authorization
    policies, and provides the ability to do attribute based access control
    (abac).

    Properties
    ----------
    filters : dict
        Key-value pairs used to filter on attributes of documents.
    """

    def __init__(self):
        self._filters = {}

    @property
    def filters(self):
        """Key-value pairs used to filter on attributes of documents."""

        return self._filters

    @filters.setter
    def filters(self, filters: dict):
        self._filters = filters



from inuits_policy_based_auth.authorization.base_authorization_policy import (
    BaseAuthorizationPolicy,
)


class OpenDataPolicy(BaseAuthorizationPolicy):
    """
    An authorization policy that allows every GET-request.
    """

    def authorize(self, policy_context, user_context, request_context):
        request = request_context.http_request
        if request.method == "GET":
            policy_context.access_verdict = True

        return policy_context



import inspect
import os
import re as regex

from functools import wraps
from inuits_policy_based_auth.authentication.base_authentication_policy import (
    BaseAuthenticationPolicy,
)
from inuits_policy_based_auth.authorization.base_authorization_policy import (
    BaseAuthorizationPolicy,
)
from inuits_policy_based_auth.contexts import PolicyContext, RequestContext, UserContext
from inuits_policy_based_auth.exceptions import (
    PolicyFactoryException,
    NoUserContextException,
    NoAuthenticationPoliciesToApplyException,
    NoAuthorizationPoliciesToApplyException,
    InvalidFallbackKey,
    NoFallbackKeySet,
)
from werkzeug.exceptions import Forbidden


class PolicyFactory:
    """
    A class used to apply policies.

    Methods
    -------
    get_user_context()
        Returns an object of type UserContext.

    register_authentication_policy(key, policy)
        Appends a policy to the list of authentication policies to be applied
        mapped to a key.

    register_authorization_policy(key, policy)
        Appends a policy to the list of authorization policies to be applied
        mapped to a key.

    apply_policies(request_context)
        Applies registered policies to determine access.

    set_fallback_key_for_policy_mapping(key)
        Sets a fallback key for policy mapping if no key matches the first part
        of the module name of endpoints decorated by the policy factory.
    """

    def __init__(self):
        self._user_context = None
        self._authentication_policies: dict[str, list[BaseAuthenticationPolicy]] = {}
        self._authorization_policies: dict[str, list[BaseAuthorizationPolicy]] = {}
        self._fallback_key_for_policy_mapping = ""

    def get_user_context(self) -> UserContext:
        """
        Returns an object of type UserContext.

        Returns
        -------
        UserContext
            An object containing data about the authenticated user.

        Raises
        ------
        NoUserContextException
            If there is no user auth data yet.
        """

        if not self._user_context:
            raise NoUserContextException()
        return self._user_context

    def register_authentication_policy(
        self, key: str, policy: BaseAuthenticationPolicy
    ):
        """
        Appends a policy to the list of authentication policies to be applied mapped
        to a key.

        Parameters
        ----------
        key : str
            A key that matches the first part of the module name of endpoints decorated
            by the policy factory.
        policy : BaseAuthenticationPolicy
            A policy to be applied.
        """

        policies = self._authentication_policies.get(key)
        if policies:
            policies.append(policy)
        else:
            self._authentication_policies.update({key: [policy]})

    def register_authorization_policy(self, key: str, policy: BaseAuthorizationPolicy):
        """
        Appends a policy to the list of authorization policies to be applied mapped
        to a key.

        Parameters
        ----------
        key : str
            A key that matches the first part of the module name of endpoints decorated
            by the policy factory.
        policy : BaseAuthorizationPolicy
            A policy to be applied.
        """

        policies = self._authorization_policies.get(key)
        if policies:
            policies.append(policy)
        else:
            self._authorization_policies.update({key: [policy]})

    def set_fallback_key_for_policy_mapping(self, key: str):
        """
        Sets a key for policy mapping to fallback if no key matches the first part of
        the module name of endpoints decorated by the policy factory.

        Parameters
        ----------
        key : str
            A fallback key for policy mapping.

        Raises
        ------
        InvalidFallbackKey
            If provided key is not registered in either authentication policies,
            authorization policies, or both.
        """

        if (
            key != ""
            and self._authentication_policies.get(key)
            and self._authorization_policies.get(key)
        ):
            self._fallback_key_for_policy_mapping = key
        else:
            raise InvalidFallbackKey(key)

    def authenticate(self):
        """
        Applies registered authentication policies to determine access.

        Returns
        -------
        function
            A decorator to decorate endpoints with.

        Raises
        ------
        NoAuthenticationPoliciesToApplyException
            If no authentication policies are registered.
        Unauthorized
            If user is not authenticated.
        NoFallbackKeySet
            If no fallback key for policy mapping is set.
        """

        def decorator(decorated_function):
            @wraps(decorated_function)
            def decorated_function_wrapper(*args, **kwargs):
                if len(self._authentication_policies) <= 0:
                    raise NoAuthenticationPoliciesToApplyException()

                if self.__is_test_environment(decorated_function):
                    import traceback
                    from flask import make_response

                    try:
                        self._user_context = self._authenticate(decorated_function)
                    except (TypeError, PolicyFactoryException) as error:
                        return make_response(
                            {
                                "message": str(error),
                                "stacktrace": traceback.format_exc(),
                            },
                            500,
                        )
                else:
                    self._user_context = self._authenticate(decorated_function)

                return decorated_function(*args, **kwargs)

            return decorated_function_wrapper

        return decorator

    def apply_policies(self, request_context: RequestContext):
        """
        Applies registered policies to determine access.

        The first allowing policy will stop execution and allow access.
        The first denying policy will stop execution and deny access.
        If all policies are applied and none of them allowed or denied access
        (policy_context.access_verdict == None), then access is denied.

        Parameters
        ----------
        request_context : RequestContext
            An object containing data about the context of a request.

        Returns
        -------
        function
            A decorator to decorate endpoints with.

        Raises
        ------
        NoAuthenticationPoliciesToApplyException
            If no authentication policies are registered.
        NoAuthorizationPoliciesToApplyException
            If no authorization policies are registered.
        Unauthorized
            If user is not authenticated.
        Forbidden
            If user is not authorized.
        NoFallbackKeySet
            If no fallback key for policy mapping is set.
        """

        def decorator(decorated_function):
            @wraps(decorated_function)
            def decorated_function_wrapper(*args, **kwargs):
                if self.__is_test_environment(decorated_function):
                    import traceback
                    from flask import make_response

                    try:
                        self._apply_policies_decorated_function_wrapper_implementation(
                            decorated_function, request_context
                        )
                    except (TypeError, PolicyFactoryException) as error:
                        return make_response(
                            {
                                "message": str(error),
                                "stacktrace": traceback.format_exc(),
                            },
                            500,
                        )
                else:
                    self._apply_policies_decorated_function_wrapper_implementation(
                        decorated_function, request_context
                    )

                return decorated_function(*args, **kwargs)

            return decorated_function_wrapper

        return decorator

    def _apply_policies_decorated_function_wrapper_implementation(
        self, decorated_function, request_context: RequestContext
    ):
        if len(self._authentication_policies) <= 0:
            raise NoAuthenticationPoliciesToApplyException()
        if len(self._authorization_policies) <= 0:
            raise NoAuthorizationPoliciesToApplyException()

        self._user_context = self._authenticate(decorated_function)
        self._authorize(decorated_function, self._user_context, request_context)

    def _authenticate(self, decorated_function):
        user_context = UserContext()

        key = self._get_key_for_policy_mapping(
            self._authentication_policies, decorated_function
        )
        for policy in self._authentication_policies[key]:
            user_context = policy.apply(user_context)

        return user_context

    def _authorize(
        self,
        decorated_function,
        user_context: UserContext,
        request_context: RequestContext,
    ):
        policy_context = PolicyContext()

        key = self._get_key_for_policy_mapping(
            self._authorization_policies, decorated_function
        )
        for policy in self._authorization_policies[key]:
            policy_context = policy.apply(policy_context, user_context, request_context)

            if policy_context.access_verdict:
                return
            elif policy_context.access_verdict == False:
                break

        raise Forbidden()

    def _get_key_for_policy_mapping(self, policy_dict: dict, decorated_function) -> str:
        for key in policy_dict.keys():
            if regex.match(rf"^{key}.*", decorated_function.__module__) != None:
                return key

        if self._fallback_key_for_policy_mapping:
            return self._fallback_key_for_policy_mapping

        raise NoFallbackKeySet

    def __is_test_environment(self, decorated_function) -> bool:
        decorated_function_module = inspect.getmodule(decorated_function)
        if decorated_function_module:
            return (
                regex.match(
                    rf".*/inuits-policy-based-auth/{os.getenv('FLASK_APP')}$",
                    str(decorated_function_module.__file__),
                )
                != None
            )

        return False



import pytest

from inuits_policy_based_auth import BaseAuthorizationPolicy
from inuits_policy_based_auth.contexts import PolicyContext, RequestContext, UserContext
from inuits_policy_based_auth.exceptions import (
    AuthorizeMethodDidNotReturnObjectOfTypePolicyContextException,
)
from unittest.mock import Mock, patch


class TestBaseAuthorizationPolicy:
    @patch(
        "inuits_policy_based_auth.authorization.base_authorization_policy.BaseAuthorizationPolicy.__abstractmethods__",
        set(),
    )
    def setup_method(self, _):
        self.policy = BaseAuthorizationPolicy()  # type: ignore
        self.spy_policy_authorize = Mock()
        self.policy_context = PolicyContext()
        self.user_context = UserContext()
        self.request_context = RequestContext(None)

        self.policy.authorize = self.spy_policy_authorize

    def test_apply_returns_policy_context(self):
        self.policy.authorize.return_value = self.policy_context

        policy_context = self.policy.apply(
            self.policy_context, self.user_context, self.request_context
        )

        self.policy.authorize.assert_called_once()
        assert policy_context is self.policy_context

    def test_apply_sets_policy_context_access_verdict_to_none_by_default(self):
        self.policy_context.access_verdict = True
        self.policy.authorize.return_value = self.policy_context

        policy_context = self.policy.apply(
            self.policy_context, self.user_context, self.request_context
        )

        self.policy.authorize.assert_called_once()
        assert policy_context.access_verdict == None

    def test_apply_raises_AuthorizeMethodDidNotReturnObjectOfTypePolicyContextException(
        self,
    ):
        with pytest.raises(
            AuthorizeMethodDidNotReturnObjectOfTypePolicyContextException
        ):
            self.policy.apply(
                self.policy_context, self.user_context, self.request_context
            )

        self.policy.authorize.assert_called_once()



import pytest

from inuits_policy_based_auth import (
    PolicyFactory,
    BaseAuthenticationPolicy,
    BaseAuthorizationPolicy,
)
from inuits_policy_based_auth.contexts import PolicyContext, RequestContext, UserContext
from inuits_policy_based_auth.exceptions import (
    NoUserContextException,
    NoAuthenticationPoliciesToApplyException,
    NoAuthorizationPoliciesToApplyException,
    InvalidFallbackKey,
    NoFallbackKeySet,
)
from unittest.mock import Mock, patch
from werkzeug.exceptions import Forbidden


class TestPolicyFactory:
    def setup_method(self):
        self.policy_factory = PolicyFactory()
        self.policy_context = PolicyContext()
        self.request_context = RequestContext(None)
        self.user_context = UserContext()
        self.key = "tests.unit"

    def test_get_user_context_does_not_raise_error_if_user_context_is_present(self):
        self.policy_factory._user_context = self.user_context
        user_context = self.policy_factory.get_user_context()
        assert user_context is self.policy_factory._user_context

    def test_get_user_context_raises_NoUserContextException_if_user_context_is_none(
        self,
    ):
        with pytest.raises(NoUserContextException):
            self.policy_factory.get_user_context()

    def test_register_authentication_policy_appends_policy_to_authentication_policies(
        self,
    ):
        dummy_policy = Mock()

        self.policy_factory.register_authentication_policy(self.key, dummy_policy)

        assert len(self.policy_factory._authorization_policies) == 0
        assert len(self.policy_factory._authentication_policies) == 1
        assert self.policy_factory._authentication_policies[self.key][0] is dummy_policy

    def test_register_authorization_policy_appends_policy_to_authorization_policies(
        self,
    ):
        dummy_policy = Mock()

        self.policy_factory.register_authorization_policy(self.key, dummy_policy)

        assert len(self.policy_factory._authentication_policies) == 0
        assert len(self.policy_factory._authorization_policies) == 1
        assert self.policy_factory._authorization_policies[self.key][0] is dummy_policy

    def test_set_fallback_key_for_policy_mapping_sets_key(self):
        self.policy_factory._authentication_policies.update({self.key: [Mock()]})
        self.policy_factory._authorization_policies.update({self.key: [Mock()]})

        self.policy_factory.set_fallback_key_for_policy_mapping(self.key)

        assert self.policy_factory._fallback_key_for_policy_mapping == self.key

    def test_set_fallback_key_for_policy_mapping_raises_InvalidFallbackKey_if_key_is_an_empty_string(
        self,
    ):
        self.policy_factory._authentication_policies.update({self.key: [Mock()]})
        self.policy_factory._authorization_policies.update({self.key: [Mock()]})

        with pytest.raises(InvalidFallbackKey):
            self.policy_factory.set_fallback_key_for_policy_mapping("")

    def test_set_fallback_key_for_policy_mapping_raises_InvalidFallbackKey_if_key_is_not_registered_in_authentication_policies(
        self,
    ):
        self.policy_factory._authentication_policies.update({"different_key": [Mock()]})
        self.policy_factory._authorization_policies.update({self.key: [Mock()]})

        with pytest.raises(InvalidFallbackKey):
            self.policy_factory.set_fallback_key_for_policy_mapping(self.key)

    def test_set_fallback_key_for_policy_mapping_raises_InvalidFallbackKey_if_key_is_not_registered_in_authorization_policies(
        self,
    ):
        self.policy_factory._authentication_policies.update({self.key: [Mock()]})
        self.policy_factory._authorization_policies.update({"different_key": [Mock()]})

        with pytest.raises(InvalidFallbackKey):
            self.policy_factory.set_fallback_key_for_policy_mapping(self.key)

    def test_set_fallback_key_for_policy_mapping_raises_InvalidFallbackKey_if_key_is_not_registered_in_authentication_and_authorization_policies(
        self,
    ):
        self.policy_factory._authentication_policies.update({"different_key": [Mock()]})
        self.policy_factory._authorization_policies.update({"different_key": [Mock()]})

        with pytest.raises(InvalidFallbackKey):
            self.policy_factory.set_fallback_key_for_policy_mapping(self.key)

    def test_apply_policies_raises_NoAuthenticationPoliciesToApplyException_if_no_authentication_policies_are_registered(
        self,
    ):
        with pytest.raises(NoAuthenticationPoliciesToApplyException):
            self.policy_factory._apply_policies_decorated_function_wrapper_implementation(
                Mock(), Mock()
            )

    def test_apply_policies_raises_NoAuthorizationPoliciesToApplyException_if_no_authorization_policies_are_registered(
        self,
    ):
        self.policy_factory._authentication_policies.update({self.key: [Mock()]})

        with pytest.raises(NoAuthorizationPoliciesToApplyException):
            self.policy_factory._apply_policies_decorated_function_wrapper_implementation(
                Mock(), Mock()
            )

    def test_apply_policies_calls_authenticate_and_authorize(self):
        spy_decorated_function = Mock()
        spy_policy_factory_authenticate = Mock()
        spy_policy_factory_authorize = Mock()

        spy_policy_factory_authenticate.return_value = self.user_context
        self.policy_factory._authenticate = spy_policy_factory_authenticate
        self.policy_factory._authorize = spy_policy_factory_authorize

        self.policy_factory._authentication_policies.update({self.key: [Mock()]})
        self.policy_factory._authorization_policies.update({self.key: [Mock()]})
        self.policy_factory._apply_policies_decorated_function_wrapper_implementation(
            spy_decorated_function, self.request_context
        )

        spy_policy_factory_authenticate.assert_called_once_with(spy_decorated_function)
        spy_policy_factory_authorize.assert_called_once_with(
            spy_decorated_function, self.user_context, self.request_context
        )

    @patch(
        "inuits_policy_based_auth.authentication.base_authentication_policy.BaseAuthenticationPolicy.__abstractmethods__",
        set(),
    )
    def test_authenticate_applies_authentication_policies(self):
        spy_policy_factory_get_key_for_policy_mapping = Mock()
        dummy_decorated_function = Mock()
        policy_1 = BaseAuthenticationPolicy()  # type: ignore
        policy_2 = BaseAuthenticationPolicy()  # type: ignore
        spy_policy_1_apply = Mock()
        spy_policy_2_apply = Mock()

        spy_policy_factory_get_key_for_policy_mapping.return_value = self.key
        self.policy_factory._get_key_for_policy_mapping = (
            spy_policy_factory_get_key_for_policy_mapping
        )
        spy_policy_1_apply.return_value = self.user_context
        spy_policy_2_apply.return_value = self.user_context
        policy_1.apply = spy_policy_1_apply
        policy_2.apply = spy_policy_2_apply
        self.policy_factory._authentication_policies.update(
            {self.key: [policy_1, policy_2]}
        )

        user_context = self.policy_factory._authenticate(dummy_decorated_function)

        spy_policy_factory_get_key_for_policy_mapping.assert_called_once_with(
            self.policy_factory._authentication_policies, dummy_decorated_function
        )
        spy_policy_1_apply.assert_called_once()
        spy_policy_2_apply.assert_called_once()
        assert user_context is self.user_context

    def test_authorize_allows_access_and_stops_execution_if_policy_context_access_verdict_is_true(
        self,
    ):
        self._prepare_test_authorize()
        self.policy_context.access_verdict = True

        self.policy_factory._authorize(
            self.dummy_decorated_function, self.user_context, self.request_context
        )

        self.spy_policy_factory_get_key_for_policy_mapping.assert_called_once_with(
            self.policy_factory._authorization_policies, self.dummy_decorated_function
        )
        self.spy_policy_1_apply.assert_called_once()
        self.spy_policy_2_apply.assert_not_called()

    def test_authorize_denies_access_and_stops_execution_if_policy_context_access_verdict_is_false(
        self,
    ):
        self._prepare_test_authorize()
        self.policy_context.access_verdict = False

        with pytest.raises(Forbidden):
            self.policy_factory._authorize(
                self.dummy_decorated_function, self.user_context, self.request_context
            )

        self.spy_policy_factory_get_key_for_policy_mapping.assert_called_once_with(
            self.policy_factory._authorization_policies, self.dummy_decorated_function
        )
        self.spy_policy_1_apply.assert_called_once()
        self.spy_policy_2_apply.assert_not_called()

    def test_authorize_denies_access_if_policy_context_access_verdict_is_none(self):
        self._prepare_test_authorize()
        self.policy_context.access_verdict = None

        with pytest.raises(Forbidden):
            self.policy_factory._authorize(
                self.dummy_decorated_function, self.user_context, self.request_context
            )

        self.spy_policy_factory_get_key_for_policy_mapping.assert_called_once_with(
            self.policy_factory._authorization_policies, self.dummy_decorated_function
        )
        self.spy_policy_1_apply.assert_called_once()
        self.spy_policy_2_apply.assert_called_once()

    def test_get_key_for_policy_mapping_gets_key(self):
        stub_decorated_function = Mock()
        policy_dict = {self.key: [], "different_key": []}

        stub_decorated_function.__module__ = f"{self.key}.example"
        key = self.policy_factory._get_key_for_policy_mapping(
            policy_dict, stub_decorated_function
        )

        assert key == self.key

    def test_get_key_for_policy_mapping_uses_fallback_key(self):
        stub_decorated_function = Mock()
        different_key = "different_key"
        policy_dict = {self.key: [], different_key: []}

        stub_decorated_function.__module__ = f"invalid_module.{self.key}.example"
        self.policy_factory._fallback_key_for_policy_mapping = different_key

        key = self.policy_factory._get_key_for_policy_mapping(
            policy_dict, stub_decorated_function
        )

        assert key == different_key

    def test_get_key_for_policy_mapping_raises_NoFallbackKeySet_if_no_fallback_key_is_set(
        self,
    ):
        stub_decorated_function = Mock()
        different_key = "different_key"
        policy_dict = {self.key: [], different_key: []}

        stub_decorated_function.__module__ = f"invalid_module.{self.key}.example"
        with pytest.raises(NoFallbackKeySet):
            self.policy_factory._get_key_for_policy_mapping(
                policy_dict, stub_decorated_function
            )

    @patch(
        "inuits_policy_based_auth.authorization.base_authorization_policy.BaseAuthorizationPolicy.__abstractmethods__",
        set(),
    )
    def _prepare_test_authorize(self):
        self.spy_policy_factory_get_key_for_policy_mapping = Mock()
        self.dummy_decorated_function = Mock()
        policy_1 = BaseAuthorizationPolicy()  # type: ignore
        policy_2 = BaseAuthorizationPolicy()  # type: ignore
        self.spy_policy_1_apply = Mock()
        self.spy_policy_2_apply = Mock()

        self.spy_policy_factory_get_key_for_policy_mapping.return_value = self.key
        self.policy_factory._get_key_for_policy_mapping = (
            self.spy_policy_factory_get_key_for_policy_mapping
        )
        self.spy_policy_1_apply.return_value = self.policy_context
        self.spy_policy_2_apply.return_value = self.policy_context
        policy_1.apply = self.spy_policy_1_apply
        policy_2.apply = self.spy_policy_2_apply
        self.policy_factory._authorization_policies.update(
            {self.key: [policy_1, policy_2]}
        )
